#!/usr/bin/perl -w

our $VERSION='1.201301060257';

=head1 NAME

UIUC_Search.pl - find an ideal UIUC airfoil for a given Re

=head1 SYNOPSIS

	perl UIUC_Search.pl [options]

or

	UIUC_Search.exe [options]

run this program from the command line in the same folder as Xfoil.exe, as above.
Get perl from here if you need it: http://www.activestate.com/activeperl/downloads

     Options:
	(none)		run without options to be prompted for everything.
       -help	-?	brief help message
       -man		full documentation
       -re   nnn	supply reynolds number (eg: 1000000)
       -ncrit n		supply Ncrit (eg: 4)
       -folder		tell program what folder name has all the airfoil .dat files in it (eg: coord_seligFmt)
       -plot		activate plotting during run (is disable by default)
       -out		name of file for the output to be written into

=head1 DESCRIPTION

Do a sort on the output - the best-performing foils will then be at the end.

=head1 SUBROUTINES

=cut

######################################################################

use strict;
use warnings;		# same as -w switch above

use Getopt::Long;	# Commandline argument parsing
use Pod::Usage;		# Inbuilt documentation helper

my %arg;		# We store copies of commandline arguments in here

# Parse options, and print help/manual if requested.
&GetOptions('help|?' => \$arg{'help'},			# breif instructions
	    'man' => \$arg{'man'},			# complete manual
	    're=i'=>\$arg{'re'},			# Re
	    'ncrit|N=i'=>\$arg{'ncrit'},		# Ncrit
	    'plot'=>\$arg{'plot'},			# Specify this to view plotting during a run
	    'folder|f=s'=>\$arg{'folder'},		# folder if airfoil .dat files to use
	    'out=s'=>\$arg{'out'},			# output file
	   ) or &pod2usage(2); 
no warnings;
	   &pod2usage(1) if ($arg{'help'});
	   &pod2usage(-exitstatus => 0, -verbose => 2) if $arg{'man'};
use warnings;


# Ask user some starting paramaters:
my $folder=$arg{'folder'}; $folder=&prompt("What folder of airfiols to use?",'coord_seligFmt') unless(defined $arg{'folder'}); die "Folder '$folder' not found" unless(-d $folder);
my $Re=$arg{'re'}; $Re=&prompt("Re=speed (m/s) / chord (m) * kinematic viscosity (1.46E-05). What Re?",1000000) unless(defined $arg{'re'});
my $NCrit=$arg{'ncrit'}; $NCrit=&prompt("Dirty Air=4, clean=9. What NCrit?",4) unless(defined $arg{'ncrit'});
my $outfile=$arg{'out'}; $outfile=&prompt("Output file name (leave blank for screen output only):") unless(defined $arg{'re'}); # re is deliberate; don't ask for file if Re was given.
my $slash='/'; $slash="\\" if($^O =~/Win32/i);
my $killed=0; my %stuck; # for killing dead xfoil processes


my $a1=-3; my $a2=67;		# Range for testing (should auto-adjust really - smaller a2 for larger Re)
my $tests=70;
my $step=($a2-$a1)/$tests;	# How many AoA's to try
my $substep=$step/5;		# How much "finer" will we re-examine the ones with the best Cl/Cd
my $subtest=20;			# How many sub re-examinations will we do?


my $tmpf="tmpcmds_$$.txt";	# temporary file name for storing Xfoil commands into
my $polf="tmppolar_$$.txt";	# temp file for polar output

# &killstuck();	fails at end and locks up in run	# Set up a parallel thread to kill dead xfoil processes

if($outfile ne '') {open(OUTF,'>',$outfile) or die "Cannot open '$outfile' for writing: $!"};
#my $out=sprintf("CL/CD\tFoil    \tAlpha\tXfoil output\n"); print $out; print OUTF $out if($outfile ne '');
my $out="# $0 v$VERSION ${\gmnow()} Re=$Re NCrit=$NCrit f=$arg{'folder'} outfile=$arg{'out'} AoA($a1..$a2)by$step (${tests}t) s=$subtest p=$^O\n"; &pprint($out);
$out=sprintf("CL/CD\tFoil    \t  alpha    CL        CD       CDp       CM     Top_Xtr  Bot_Xtr\n"); &pprint($out);


  opendir(my $THISDIR,$folder) or die "can't open '$folder': $! ";
  while(my $file=readdir $THISDIR) {
    next if (($file eq '.')||($file eq '..'));

	my $xfcmd="\n\n";	# Accept any defaults/questions it might ask at start
	$xfcmd.="NORM\n";	# Turn on normalization
	$xfcmd.="PLOP\nG\n\n" unless($arg{'plot'});	# Turn off graphix plotting
	$xfcmd.="LOAD $folder$slash$file\n";	# Load the foil
	#$xfcmd.="MDES\nFILT\nEXEC\n\n\nPCOP\n";	# Smooth any variations in the dat file data
	$xfcmd.="GDES\nCADD\n\n\n\n\nPCOP\n";	# smoother panelling
	$xfcmd.="PANE\n";	# Fix possible problems it might have
	$xfcmd.="PCOP\n";	# Copy buffer back
	$xfcmd.="OPER\n";	# Go into operator menu
	$xfcmd.="ITER 75\n";	# Allow working harder to find convergance
	$xfcmd.="VPAR\n";	# So we can set Ncrit
	$xfcmd.="N $NCrit\n";	# set it
	$xfcmd.="\n";		# exit VPAR submenu
	$xfcmd.="VISC $Re\n";	# set Re
	$xfcmd.="P\n";		# save polars to a file
	$xfcmd.="$polf\n";	# this one
	$xfcmd.="\n";		# no dumps thanks
	
	my $fine=$xfcmd;
        for(my $alfa=$a1;$alfa<$a2;$alfa+=$step) {
	  $xfcmd.="A $alfa\nINIT\n";	# test it (needs init after, otherwise results get bogus)
        } # alfa

	$xfcmd.="\n";		# exit submenu
	$xfcmd.="\n\nQUIT\n\n\n";# end

	my @pol=&runxfoil($xfcmd);	# Rough search

	my ($bestld,$besta,$nam)=&procpol(@pol);

	$out="# SubBest Cl/Cd=$bestld at AoA=$besta\t$file\t$nam\n"; &pprint($out);

	if($besta!~/\d/) {	# found none!
	  $besta=4; # pick a default to re-try - maybe just xfoil crashed out
	}


	$xfcmd=$fine;	# re-init XFOIL with same parms
        my $alfa=$besta-(($subtest*$substep)/2); my $loop=$subtest;
        while($loop--) {
	  $xfcmd.="A $alfa\n";	# test it
	  $alfa+=$substep;
        } # alfa

	$xfcmd.="\n";		# exit submenu
	$xfcmd.="\n\nQUIT\n\n\n";# end

	@pol=&runxfoil($xfcmd);		# Fine search

	($bestld,$besta,$nam)=&procpol(@pol);
	next if($besta!~/\d/);	# found none!

	$out="# Best Cl/Cd=$bestld at AoA=$besta\t$file\t$nam\n"; &pprint($out);

  } # while


$out="UIUC_Search run complete.\n"; print $out; print OUTF $out if($outfile ne '');
close OUTF;

#
# Invoke xfoil and get back data from it
#
sub runxfoil {
  my($xfcmd)=@_;

  &writefile($tmpf,$xfcmd);

  unlink($polf) if(-e $polf); die "Cannot erase existing file $polf: $!" if(-e $polf);
  my $res=`xfoil <$tmpf`; # This runs Xfoil.  If it locks up, use task manager to kill xfoil.exe, and it will resume.
  my @pol;
  if(-e $polf) {
    open(IN,'<',$polf) or die "Cannot read '$polf': $!";
    @pol=<IN>; close(IN);	# slurp it all
  } else {
    warn "xfoil failed to produce $polf!";
    warn $res;
  }
  return @pol;
} #runxfoil



#
# Process @pol and find best AoA (optimal CL/CD)
#
sub procpol {
  my(@pol)=@_;

  my($bestld,$besta,$nam); my $go=0;
  
  foreach(@pol) {
    if((!defined $nam)&&(/Calculated polar for: (.*)/)) {$nam=$1; $nam=~s/\s+$//;}
    if(/--$/){$go++; next;}
    next unless($go);

    my(undef,$alpha,$CL,$CD,$CDp,$CM,$Top_Xtr,$Bot_Xtr)=split(/\s+/," $_");
    next if($CD=~/(NAN|^$)/);
    next if($CD<=0);	# nothing/nothing sensible
    next if($CDp<=0);	# non-convergance data muckup
    my $clcd=($CL/$CD);
    ($bestld,$besta)=($clcd,$alpha) if((!defined $bestld)||($clcd>$bestld));
    $clcd=sprintf("%.02f",$clcd);		# round to 2 decimal places
    $out=sprintf("%6s\t$nam\t$_",$clcd);	# space-pad clcd for sorting
    &pprint($out);
  }
  return ($bestld,$besta,$nam);
} # procpol



#
# Print something on screen (and outfile)
#

sub pprint {
  my($out)=@_;
  print $out; print OUTF $out if($outfile ne '');
} # pprint



#
# Set up a parallel thread to kill dead xfoil processes
#
sub killstuck {
  my $pid = fork;
  if ($pid == 0) {		# child
    while(1) {
      sleep(30); my $xp='';
      if($^O =~/Win32/i) {

	my @dat=`wmic process list`;
	my @cols=split(/\s+/,$dat[0]);
	my($i,$idx); foreach $i (0..$#cols) { if($cols[$i] eq 'ProcessId'){$idx=(($#cols)-$i);last}; }
	if(defined $idx) {
	  foreach(@dat) {
	    if(/\sxfoil.exe\s/) {
	      @cols=split(/\s+/,$_);
	      $xp.=" " . $cols[($#cols)-$idx+2];
	    }
	  }
	} # idx
      } else {
        $xp=`pidof xfoil`;	# Linux
      }

      my %new;
      my @pids=split(/\s+/,$xp);
      foreach my $pid (@pids) { next if($pid!~/\d/);
	if($stuck{$pid}++>5) {
	  $stuck{$pid}=0;
	  $killed++;
	  print "killing $pid ($killed)\n";
	  if($^O =~/Win32/i) {
            `taskkill /F /PID $pid`;
	  } else {
	    `kill $pid`;		# Linux
	  }
	} # stuck
	$new{$pid}=$stuck{$pid};
      } # pids

      %stuck=%new;	# Discard old non-stuck stuff

    } # 1
    exit(0);
  } # child
} # killstuck



#
# Write to a file, checking we have OK permissions
#
sub writefile {
  my($file,$msg)=@_;
  unlink($file) if(-e $file);
  die "Cannot erase existing file $file: $!" if(-e $file);
  open(OUT,'>',$file) or die "Cannot open file '$file' for writing: $!";
  print OUT $msg;
  close(OUT);
} # writefile



#
# Ask the user something
#
sub prompt {
  my($msg,$dflt)=@_; print "$msg";
  if(defined $dflt) { print " [$dflt] "; } # show the default
  my $reply=<STDIN>; chomp $reply;
  if((defined $dflt)&&($reply eq '')) { $reply=$dflt; print "$reply\n"; }
  return $reply;
} # prompt


sub gmnow {
  my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime(time);
  return sprintf("%02d/%02d %02d:%02d.%02d",$mon+1,$mday,$hour,$min,$sec);
}


=head1 AUTHOR

C. N. Drake, E<lt>christopher@pobox.comE<gt>

=head1 WARRANTY and DISCLAIMER

Important. There is no warranty (see License below).

USE AT OWN RISK!

=head1 COPYRIGHT AND LICENSE

This is distrubed under the GNU Public License:-
http://www.gnu.org/copyleft/gpl.html

Copyright (C) 2013 by C. N. Drake.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.3 or,
at your option, any later version of Perl 5 you may have available.

=cut
